import ast
import contextlib
import json
import sys
from functools import lru_cache
from io import StringIO
from typing import Optional

import pandas as pd
from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

from ...dataset.manager import Manager


class Query(BaseModel):
    code: str
    start: Optional[int] = None
    end: Optional[int] = None
    byImage: bool = False
    cached: bool = False


@contextlib.contextmanager
def capture():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


def build_queries_router(params, manager: Manager, meta_serving: bool = False):
    queries_router = r = APIRouter(tags=["Dataset: Dataframe queries"])

    df = manager.anno_df.copy()

    if manager.task in ["classification", "multilabel_classification"]:
        df["type"] = "image"

    elif manager.task in ["detection", "box_classification", "box_multilabel_classification"]:
        df["type"] = "box"

    elif manager.task == "instance_segmentation":
        df["type"] = "mask"

    elif manager.task == "keypoints_detection":
        df["type"] = "keypoints"

    gdict = {"pd": pd, "ast": ast}
    ldict = {"df": df}

    meta_df = None
    if meta_serving:
        meta_df = manager.meta_df.copy()
        meta_df.rename(columns={"file_name": "image_hash"}, inplace=True)
        ldict["meta_df"] = meta_df

    @lru_cache(maxsize=5)
    def run_query(code: str):
        with capture() as out:
            try:
                exec(code, gdict, ldict)
            except Exception as e:
                raise HTTPException(status_code=409, detail=str(e))

            res = ldict.get("res")

            if res is None:
                raise HTTPException(status_code=409, detail="res variable is not found in query code")

            if isinstance(res, pd.Series):
                res = res.to_frame("data")

            if isinstance(res, pd.DataFrame):
                if res.index.names[0] is not None:
                    res = res.reset_index()

                return {"log": out[0].getvalue(), "res": res}

            raise HTTPException(status_code=409, detail="res variable is not DataFrame or Series")

    @r.post("/queries", summary="Query the annotation dataframe")
    async def query_r(query: Query):
        if query.cached:
            res = run_query(query.code)
        else:
            res = run_query.__wrapped__(query.code)

        res_df = res["res"]
        if query.byImage and "image_hash" in res_df.columns:
            hashes = res_df["image_hash"].unique()
            start = query.start or 0
            end = len(hashes) if query.end is None else query.end
            size = len(hashes)
            res_slice = res_df[res_df["image_hash"].isin(hashes[start:end])]
        else:
            start = query.start or 0
            end = len(res_df) if query.end is None else query.end
            size = len(res_df)
            res_slice = res_df[start:end]

        return {
            "log": res["log"],
            "result": {
                "header": res_slice.columns.values.tolist(),
                "data": json.loads(res_slice.to_json(orient="values")),
                "size": size,
            },
        }

    return queries_router, df, meta_df



import logging
import os
from pathlib import Path

from rich.logging import RichHandler

ZITRC_FILE = Path(os.path.expanduser("~/.zitrc"))
REGISTRY_ENDPOINT = "https://www.api.zityspace.cn/registry"
WS_FORMULA_PUBLISH_ENDPOINT = "wss://www.api.zityspace.cn/registry/formula/publish"
WS_FORMULA_INSTALL_ENDPOINT = "wss://www.api.zityspace.cn/registry/formula/install"
AUTH_PUBLIC_ENDPOINT = "https://www.api.zityspace.cn/auth/public"


logging.basicConfig(level="INFO", format="%(message)s", handlers=[RichHandler(rich_tracebacks=True)])
logger = logging.getLogger(__name__)

logging.getLogger("httpx").setLevel(logging.WARNING)



import os
from enum import Enum
from pathlib import Path

import httpx
import typer
from rich.progress import BarColumn, Progress, TextColumn, TimeRemainingColumn

from .config import REGISTRY_ENDPOINT, ZITRC_FILE, logger
from .utils import SubprocessError, find_zit_root, run_subprocess

example_app = app = typer.Typer(name="example", help="Example commands")


class ExampleName(str, Enum):
    detection_fashion = "detection/fashion"


# Helper function to extract filename from Content-Disposition
def get_filename_from_content_disposition(content_disposition):
    parts = content_disposition.split(";")
    for part in parts:
        if "filename=" in part:
            return part.split("=")[1].strip(' "')
    return None  # default if not found


@app.command(name="pull", help="Pull example")
def pull(name: ExampleName, unzip: bool = True):
    if not ZITRC_FILE.exists():
        logger.warning('Key not found, please login first by "zit auth login"')
        return

    with ZITRC_FILE.open("r") as f:
        token = f.read()

    zit_root = find_zit_root(Path.cwd())

    with httpx.Client(headers={"Token": token}) as client:
        with client.stream("GET", f"{REGISTRY_ENDPOINT}/example?name={name.value}") as response:
            if response.status_code == 401:
                logger.warning('Invalid key, please login first by "zit auth login"')
                return

            if response.status_code == 404:
                logger.warning(f"Example {name.value} not found")
                return

            total = int(response.headers["Content-Length"])

            content_disposition = response.headers.get("Content-Disposition", "")
            filename = get_filename_from_content_disposition(content_disposition)

            chunk_size = 81920

            with Progress(
                TextColumn(" " * 28),
                TextColumn("   receiving:"),
                BarColumn(),
                TextColumn("[yellow]{task.description}"),
                TextColumn("•"),
                TextColumn("[magenta]{task.percentage:>3.0f}%"),
                TextColumn("•"),
                TextColumn("[green]{task.completed}/{task.total}"),
                TextColumn("•"),
                TimeRemainingColumn(),
            ) as progress:
                task = progress.add_task(description=" " * 30, total=total)

                progress.update(
                    task, description=f"{filename if len(filename) <= 30 else ('...' + filename[-27:]):<30}"
                )

                with open(zit_root / filename, "wb") as f:
                    for chunk in response.iter_bytes(chunk_size):
                        f.write(chunk)
                        progress.update(task, advance=len(chunk))

            if not unzip:
                return

            try:
                run_subprocess(["tar", "-xzf", filename], cwd=zit_root)
                logger.info("Example extracted successfully")
            except SubprocessError as e:
                logger.error(f"Error extracting example: {e}")

            os.remove(zit_root / filename)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zit', 'zit.dataset', 'zit.routes', 'zit.routes.dataset']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'duckdb>=0.8.1,<0.9.0',
 'fastapi>=0.95.1,<0.96.0',
 'httpx[socks]>=0.24.1,<0.25.0',
 'pandas>=2.0.2,<3.0.0',
 'pillow>=9.5.0,<10.0.0',
 'python-multipart>=0.0.6,<0.0.7',
 'typer[all]>=0.7.0,<0.8.0',
 'websocket-client>=1.5.1,<2.0.0']

entry_points = \
{'console_scripts': ['zit = zit.main:app']}

setup_kwargs = {
    'name': 'zit',
    'version': '0.0.5a3',
    'description': 'ZitySpace CLI tool',
    'long_description': 'None',
    'author': 'Rui Zheng',
    'author_email': 'rui@zityspace.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import httpx
import typer

from .config import AUTH_PUBLIC_ENDPOINT, ZITRC_FILE, logger

auth_app = app = typer.Typer(name="auth", help="Auth commands")


def authenticate_r(token: str = None):
    if not token:
        if not ZITRC_FILE.exists():
            return False

        with ZITRC_FILE.open("r") as f:
            token = f.read()

    with httpx.Client() as client:
        try:
            r = client.get(
                f"{AUTH_PUBLIC_ENDPOINT}/whoami",
                headers={"Authorization": f"Bearer {token}"},
            )
        except httpx.ConnectError:
            return False

        if r.status_code != 200:
            return False

        user = r.json()
        return user


def verification_code_r(email: str):
    with httpx.Client() as client:
        try:
            r = client.get(
                f"{AUTH_PUBLIC_ENDPOINT}/verification/email?email={email}&expect_registered=false", timeout=20
            )
        except httpx.ConnectError:
            return False

        if r.status_code != 200:
            if r.status_code == 409:
                logger.warning(r.json().get("detail"))

            return False

        token = r.json().get("token")
        return token


def signup_r(email: str, password: str, code: str, token: str):
    with httpx.Client() as client:
        try:
            r = client.post(
                f"{AUTH_PUBLIC_ENDPOINT}/signup/email",
                data={"email": email, "password": password, "code": code, "token": token},
                timeout=20,
            )
        except httpx.ConnectError:
            return False

        if r.status_code != 200:
            if r.status_code == 409:
                logger.warning(r.json().get("detail"))

            return False

        token = r.json().get("access_token")
        return token


def login_r(email: str, password: str):
    with httpx.Client() as client:
        try:
            r = client.post(
                f"{AUTH_PUBLIC_ENDPOINT}/token/email",
                data={"email": email, "password": password},
                timeout=20,
            )
        except httpx.ConnectError:
            return False

        if r.status_code != 200:
            if r.status_code == 401:
                logger.warning(r.json().get("detail"))

            return False

        token = r.json().get("access_token")
        return token


@app.command(name="signup", help="Zit signup")
def signup():
    logger.info("Please go to https://zityspace.cn/signup for signing up. Then run 'zit auth login' to login.")

    # email = typer.prompt("Enter your email", hide_input=False)
    # password = typer.prompt("Enter your password", hide_input=True)

    # token = verification_code_r(email)
    # if not token:
    #     logger.info("Signup failed. Please check your email and try again.")
    #     return

    # code = typer.prompt("Please enter the verification code sent to your email", hide_input=False)

    # token = signup_r(email, password, code, token)
    # if not token:
    #     logger.info("Signup failed. Please check your email and try again.")
    #     return

    # # Store the token in the .zitrc file
    # with ZITRC_FILE.open("w") as f:
    #     f.write(token)

    # # Set the file permissions to be readable and writable only by the owner
    # ZITRC_FILE.chmod(0o600)

    # logger.info("Signup successfully.")


@app.command(name="login", help="Zit login")
def login():
    email = typer.prompt("Enter your email", hide_input=False)
    password = typer.prompt("Enter your password", hide_input=True)

    token = login_r(email, password)
    if not token:
        logger.info("Login failed. Please check your email and password.")
        return

    # Store the API key in the .zitrc file
    with ZITRC_FILE.open("w") as f:
        f.write(token)

    # Set the file permissions to be readable and writable only by the owner
    ZITRC_FILE.chmod(0o600)

    logger.info("Login successfully.")

    whoami()


@app.command(name="logout", help="Zit logout")
def logout():
    if ZITRC_FILE.exists():
        ZITRC_FILE.unlink()  # Delete the .zitrc file
        logger.info("Logout successful. Your API key has been removed.")
    else:
        logger.info("You are not currently logged in.")


@app.command(name="whoami", help="Zit whoami")
def whoami():
    if not ZITRC_FILE.exists():
        logger.info('You are not currently logged in. Please login first by "zit auth login"')
        return

    user = authenticate_r()
    if not user:
        logger.info('Token expired or invalid. Please login by "zit auth login"')
        return

    logger.info({"username": user["username"], "email": user["email"]})

    if not user["username_confirmed"]:
        logger.info(
            "Your username is auto-generated during signup. Please go to https://zityspace.cn/set-username to set your"
            " username."
        )
